"""
This module contains the classes needed for keeping realtime market data 
"""
import asyncio
import threading
from typing import Callable, Awaitable
from datetime import datetime
from tse_utils.models.instrument import Instrument, InstrumentIdentification
from tse_utils.models.realtime import OrderBookRow, ClientType
from tse_utils.tsetmc import MarketWatchTradeData, MarketWatchClientTypeData


class MarketRealtimeData:
    """Holds all realtime data for market"""

    def __init__(self):
        self.__instruments: list[Instrument] = []
        self.__instruments_lock: threading.Lock = threading.Lock()
        self.pusher_trade_data: Callable[
            [list[Instrument]], Awaitable[None]
        ] = lambda x: asyncio.sleep(0)
        self.pusher_orderbook_data: Callable[
            [list[tuple[Instrument, list[int]]]], Awaitable[None]
        ] = lambda x: asyncio.sleep(0)
        self.pusher_clienttype_data: Callable[
            [list[Instrument]], Awaitable[None]
        ] = lambda x: asyncio.sleep(0)

    def apply_new_client_type(
        self, client_type: list[MarketWatchClientTypeData]
    ) -> None:
        """Applies the new client type to the repository"""
        updated_clienttype_instruments = []
        with self.__instruments_lock:
            for mwi in client_type:
                instrument = next(
                    (
                        x
                        for x in self.__instruments
                        if x.identification.tsetmc_code == mwi.tsetmc_code
                    ),
                    None,
                )
                if instrument and instrument.client_type != mwi:
                    self.update_instrument_client_type(instrument.client_type, mwi)
                    updated_clienttype_instruments.append(instrument)
        threading.Thread(
            target=asyncio.run,
            args=(self.pusher_clienttype_data(updated_clienttype_instruments),),
            daemon=True,
        ).start()

    def update_instrument_client_type(
        self, instrument_ct: ClientType, mwi_ct: ClientType
    ) -> None:
        """Update an instrument's client type data"""
        instrument_ct.legal.buy.num = mwi_ct.legal.buy.num
        instrument_ct.legal.buy.volume = mwi_ct.legal.buy.volume
        instrument_ct.legal.sell.num = mwi_ct.legal.sell.num
        instrument_ct.legal.sell.volume = mwi_ct.legal.sell.volume
        instrument_ct.natural.buy.num = mwi_ct.natural.buy.num
        instrument_ct.natural.buy.volume = mwi_ct.natural.buy.volume
        instrument_ct.natural.sell.num = mwi_ct.natural.sell.num
        instrument_ct.natural.sell.volume = mwi_ct.natural.sell.volume

    def apply_new_trade_data(self, trade_data: list[MarketWatchTradeData]) -> None:
        """Applies the new trade data to the repository"""
        updated_trade_instruments = []
        updated_orderbook_instruments = []
        with self.__instruments_lock:
            for mwi in trade_data:
                instrument = next(
                    (
                        x
                        for x in self.__instruments
                        if x.identification.isin == mwi.identification.isin
                    ),
                    None,
                )
                if not instrument:
                    instrument = Instrument(
                        InstrumentIdentification(
                            isin=mwi.identification.isin,
                            tsetmc_code=mwi.identification.tsetmc_code,
                            ticker=mwi.identification.ticker,
                            name_persian=mwi.identification.name_persian,
                        )
                    )
                    self.__instruments.append(instrument)
                if not (
                    instrument.intraday_trade_candle.last_trade_datetime
                    and instrument.intraday_trade_candle.last_trade_datetime.time()
                    == mwi.last_trade_time
                ):
                    self.update_instrument_trade_data(instrument, mwi)
                    updated_trade_instruments.append(instrument)
                updated_rows = []
                for rn, row in enumerate(mwi.orderbook.rows):
                    if row != instrument.orderbook.rows[rn]:
                        self.update_instrument_orderbook_row(
                            instrument.orderbook.rows[rn], row
                        )
                        updated_rows.append(rn)
                if updated_rows:
                    updated_orderbook_instruments.append((instrument, updated_rows))
        threading.Thread(
            target=asyncio.run,
            args=(self.pusher_trade_data(updated_trade_instruments),),
            daemon=True,
        ).start()
        threading.Thread(
            target=asyncio.run,
            args=(self.pusher_orderbook_data(updated_orderbook_instruments),),
            daemon=True,
        ).start()

    def update_instrument_orderbook_row(
        self, instrument_obr: OrderBookRow, mwi_obr: OrderBookRow
    ) -> None:
        """Update a single row in instrument's order book"""
        instrument_obr.demand.num = mwi_obr.demand.num
        instrument_obr.demand.volume = mwi_obr.demand.volume
        instrument_obr.demand.price = mwi_obr.demand.price
        instrument_obr.supply.num = mwi_obr.supply.num
        instrument_obr.supply.volume = mwi_obr.supply.volume
        instrument_obr.supply.price = mwi_obr.supply.price

    def update_instrument_trade_data(
        self, instrument: Instrument, mwi: MarketWatchTradeData
    ) -> None:
        """Updates trade data for a single instrument"""
        instrument.order_limitations.max_price = mwi.price_thresholds.max_price
        instrument.order_limitations.min_price = mwi.price_thresholds.min_price
        instrument.intraday_trade_candle.previous_price = (
            mwi.intraday_trade_candle.previous_price
        )
        instrument.intraday_trade_candle.close_price = (
            mwi.intraday_trade_candle.close_price
        )
        instrument.intraday_trade_candle.last_price = (
            mwi.intraday_trade_candle.last_price
        )
        instrument.intraday_trade_candle.open_price = (
            mwi.intraday_trade_candle.open_price
        )
        instrument.intraday_trade_candle.max_price = mwi.intraday_trade_candle.max_price
        instrument.intraday_trade_candle.min_price = mwi.intraday_trade_candle.min_price
        instrument.intraday_trade_candle.trade_volume = (
            mwi.intraday_trade_candle.trade_volume
        )
        instrument.intraday_trade_candle.trade_value = (
            mwi.intraday_trade_candle.trade_value
        )
        instrument.intraday_trade_candle.trade_num = mwi.intraday_trade_candle.trade_num
        instrument.intraday_trade_candle.last_trade_datetime = datetime.combine(
            datetime.today(), mwi.last_trade_time
        )

    def get_instruments(self, isins: list[str]) -> list[Instrument]:
        """Returns instruments matching with a list of isins"""
        with self.__instruments_lock:
            instruments = [
                next(
                    (x for x in self.__instruments if x.identification.isin == y), None
                )
                for y in isins
            ]
        return instruments

    def get_all_instruments(self) -> list[Instrument]:
        """Returns all instruments"""
        with self.__instruments_lock:
            instruments = list(self.__instruments)
        return instruments



"""
This module contains the necessary codes for the TSETMC pusher's client.
"""
import json
import logging
from datetime import datetime
from websockets import client
from websockets.sync.client import ClientConnection
from tse_utils.models.instrument import Instrument


class TsetmcClient:
    """
The class used for connecting to the TSETMC pusher websocket \
and subscribe to its realtime data
    """

    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        websocket_host: str,
        websocket_port: int,
        subscribed_instruments: list[Instrument],
    ):
        self.websocket_host: str = websocket_host
        self.websocket_port: int = websocket_port
        self.websocket: ClientConnection = None
        self.subscribed_instruments: list[Instrument] = subscribed_instruments

    async def listen(self) -> None:
        """Listens to websocket updates"""
        while True:
            message = await self.websocket.recv()
            self._LOGGER.debug("Client received: %s", message)
            self.process_message(message=message)

    def process_message(self, message: str) -> None:
        """Processes a new message received from websocket"""
        message_js = json.loads(message)
        for isin, channels in message_js.items():
            instrument = next(
                x for x in self.subscribed_instruments if x.identification.isin == isin
            )
            for channel, data in channels.items():
                match channel:
                    case "thresholds":
                        self.__message_thresholds(instrument, data)
                    case "trade":
                        self.__message_trade(instrument, data)
                    case "orderbook":
                        self.__message_orderbook(instrument, data)
                    case "clienttype":
                        self.__message_clienttype(instrument, data)
                    case _:
                        self._LOGGER.fatal("Unknown message channel: %s", channel)

    def __message_thresholds(self, instrument: Instrument, data: list) -> None:
        """Handles a threshold update message"""
        instrument.order_limitations.max_price = int(data[0])
        instrument.order_limitations.min_price = int(data[1])

    def __message_trade(self, instrument: Instrument, data: list) -> None:
        """Handles a trade update message"""
        instrument.intraday_trade_candle.close_price = int(data[0])
        instrument.intraday_trade_candle.last_price = int(data[1])
        instrument.intraday_trade_candle.last_trade_datetime = datetime.fromisoformat(
            data[2]
        )
        instrument.intraday_trade_candle.max_price = int(data[3])
        instrument.intraday_trade_candle.min_price = int(data[4])
        instrument.intraday_trade_candle.open_price = int(data[5])
        instrument.intraday_trade_candle.previous_price = int(data[6])
        instrument.intraday_trade_candle.trade_num = int(data[7])
        instrument.intraday_trade_candle.trade_value = int(data[8])
        instrument.intraday_trade_candle.trade_volume = int(data[9])

    def __message_orderbook(self, instrument: Instrument, data: list) -> None:
        """Handles an orderbook update message"""
        for row in data:
            rn = int(row[0])
            instrument.orderbook.rows[rn].demand.num = int(row[1])
            instrument.orderbook.rows[rn].demand.price = int(row[2])
            instrument.orderbook.rows[rn].demand.volume = int(row[3])
            instrument.orderbook.rows[rn].supply.num = int(row[4])
            instrument.orderbook.rows[rn].supply.price = int(row[5])
            instrument.orderbook.rows[rn].supply.volume = int(row[6])

    def __message_clienttype(self, instrument: Instrument, data: list) -> None:
        """Handles an orderbook update message"""
        instrument.client_type.legal.buy.num = int(data[0])
        instrument.client_type.legal.buy.volume = int(data[1])
        instrument.client_type.legal.sell.num = int(data[2])
        instrument.client_type.legal.sell.volume = int(data[3])
        instrument.client_type.natural.buy.num = int(data[4])
        instrument.client_type.natural.buy.volume = int(data[5])
        instrument.client_type.natural.sell.num = int(data[6])
        instrument.client_type.natural.sell.volume = int(data[7])

    async def subscribe(self) -> None:
        """Subscribe to the channels for the appointed instruemtns"""
        self._LOGGER.info(
            "Client is subscribing to data for %d instruments.",
            len(self.subscribed_instruments),
        )
        isins = ",".join([x.identification.isin for x in self.subscribed_instruments])
        await self.websocket.send(f"1.all.{isins}")

    async def operate(self) -> None:
        """Start connecting to the websocket and listening for updates"""
        self._LOGGER.info("Client is starting its operation.")
        async with client.connect(
            f"ws://{self.websocket_host}:{self.websocket_port}"
        ) as self.websocket:
            await self.subscribe()
            await self.listen()



"""
This module contains the operational classes in this project 
"""

import asyncio
import logging
from datetime import datetime
import httpx
from tse_utils import tsetmc
from tse_utils.tsetmc.models import TsetmcScrapeException
from tsetmc_pusher.server.repository import MarketRealtimeData
from tsetmc_pusher.server.websocket import TsetmcWebsocket
from tsetmc_pusher.server.timing import (
    sleep_until,
    MARKET_END_TIME,
    MARKET_START_TIME,
    CRAWL_SLEEP_SECONDS,
    TRADE_DATA_TIMEOUT_MAX,
    TRADE_DATA_TIMEOUT_MIN,
    TRADE_DATA_TIMEOUT_STEP,
    CLIENT_TYPE_TIMEOUT_MAX,
    CLIENT_TYPE_TIMEOUT_MIN,
    CLIENT_TYPE_TIMEOUT_STEP,
)


class TsetmcOperator:
    """This module is responsible for continuously crawling TSETMC"""

    _LOGGER = logging.getLogger(__name__)

    def __init__(self, websocket_host: str, websocket_port: int):
        self.market_realtime_date: MarketRealtimeData = MarketRealtimeData()
        self.websocket = TsetmcWebsocket(
            market_realtime_data=self.market_realtime_date,
            websocket_host=websocket_host,
            websocket_port=websocket_port,
        )
        self.__trade_data_timeout: float = TRADE_DATA_TIMEOUT_MIN
        self.__client_type_timeout: float = CLIENT_TYPE_TIMEOUT_MIN
        self.__tsetmc_scraper = tsetmc.TsetmcScraper()

    async def __update_trade_data(self) -> None:
        """Updates trade data from TSETMC"""
        self._LOGGER.info(
            "Trade data catch started, timeout: %.2f", self.__trade_data_timeout
        )
        trade_data = await self.__tsetmc_scraper.get_market_watch(
            # The following line has been removed because of a bug in TSETMC server \
            # that ignores updates on some instruments, including options
            # h_even=self.__max_trade_time_int, ref_id=self.__max_order_row_id
            timeout=self.__trade_data_timeout
        )
        if trade_data:
            self.market_realtime_date.apply_new_trade_data(trade_data)

    @classmethod
    def next_market_watch_request_ids(cls, trade_data) -> tuple[int, int]:
        """Extracts the maximum trade time and orderbook row id"""
        max_order_row_id = max(
            max(y.row_id for y in x.orderbook.rows) for x in trade_data if x.orderbook
        )
        max_trade_time = max(x.last_trade_time for x in trade_data)
        max_trade_time_int = (
            max_trade_time.hour * 10000
            + max_trade_time.minute * 100
            + max_trade_time.second
        )
        return max_trade_time_int, max_order_row_id

    async def __perform_trade_data_loop(self) -> None:
        """Perform the trade data tasks for the market open time"""
        while datetime.now().time() < MARKET_END_TIME:
            try:
                await self.__update_trade_data()
                await asyncio.sleep(CRAWL_SLEEP_SECONDS)
                self.__trade_data_timeout = max(
                    TRADE_DATA_TIMEOUT_MIN,
                    self.__trade_data_timeout - TRADE_DATA_TIMEOUT_STEP,
                )
            except (
                ValueError,
                TsetmcScrapeException,
                httpx.RemoteProtocolError,
                httpx.ReadError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.ConnectTimeout,
            ) as ex:
                self._LOGGER.error("Exception on catching trade data: %s", repr(ex))
                self.__trade_data_timeout = min(
                    TRADE_DATA_TIMEOUT_MAX,
                    self.__trade_data_timeout + TRADE_DATA_TIMEOUT_STEP,
                )

    async def __update_client_type(self) -> None:
        """Updates client type from TSETMC"""
        self._LOGGER.info(
            "Client type catch started, timeout: %.2f", self.__client_type_timeout
        )
        client_type = await self.__tsetmc_scraper.get_client_type_all(
            timeout=self.__client_type_timeout
        )
        if client_type:
            self.market_realtime_date.apply_new_client_type(client_type)

    async def __perform_client_type_loop(self) -> None:
        """Perform the client type tasks for the market open time"""
        while datetime.now().time() < MARKET_END_TIME:
            try:
                await self.__update_client_type()
                await asyncio.sleep(CRAWL_SLEEP_SECONDS)
                self.__client_type_timeout = max(
                    CLIENT_TYPE_TIMEOUT_MIN,
                    self.__client_type_timeout - CLIENT_TYPE_TIMEOUT_STEP,
                )
            except (
                ValueError,
                TsetmcScrapeException,
                httpx.RemoteProtocolError,
                httpx.ReadError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.ConnectTimeout,
            ) as ex:
                self._LOGGER.error("Exception on catching client type: %s", repr(ex))
                self.__client_type_timeout = min(
                    CLIENT_TYPE_TIMEOUT_MAX,
                    self.__client_type_timeout + CLIENT_TYPE_TIMEOUT_STEP,
                )

    async def market_time_operations(self) -> None:
        """Groups the different market time operations"""
        group = asyncio.gather(
            self.__perform_trade_data_loop(),
            self.__perform_client_type_loop(),
            self.websocket.serve_websocket(),
        )
        await asyncio.wait_for(group, timeout=None)

    async def perform_daily(self) -> None:
        """Daily tasks for the crawler are called from here"""
        self._LOGGER.info("Daily tasks are starting.")
        await sleep_until(MARKET_START_TIME)
        self._LOGGER.info("Market is starting.")
        await self.market_time_operations()



"""This modules holds the necessary timing parameters for the project's operations"""
import asyncio
from datetime import time, datetime


MARKET_START_TIME: time = time(hour=8, minute=30, second=0)
MARKET_END_TIME: time = time(hour=15, minute=0, second=0)
CRAWL_SLEEP_SECONDS: float = 0.5
TRADE_DATA_TIMEOUT_MAX: float = 1.5
TRADE_DATA_TIMEOUT_MIN: float = 0.5
TRADE_DATA_TIMEOUT_STEP: float = 0.25
CLIENT_TYPE_TIMEOUT_MAX: float = 3.0
CLIENT_TYPE_TIMEOUT_MIN: float = 0.5
CLIENT_TYPE_TIMEOUT_STEP: float = 0.25


async def sleep_until(wakeup_at: time) -> None:
    """Sleep until appointed time"""
    timedelta = datetime.combine(datetime.today(), wakeup_at) - datetime.now()
    await asyncio.sleep(timedelta.total_seconds())
